import model.tasks as ts
import random

def truncate(minval, maxval):
    def _limit(fun):
        def _f(*args, **kargs):
            val = fun(*args, **kargs)
            return min(maxval, max(minval, val))
        return _f
    return _limit

def redraw(minval, maxval):
    def _redraw(dist):
        def _f(*args, **kargs):
            in_range = False
            while not in_range:
                val = dist(*args, **kargs)
                in_range = minval <= val <= maxval
            return val
        return _f
    return _redraw

def uniform_int(minval, maxval):
    def _draw():
        return random.randint(minval, maxval)
    return _draw

def uniform(minval, maxval):
    def _draw():
        return random.uniform(minval, maxval)
    return _draw

def uniform_choice(set):
    selector = uniform_int(0, len(set) - 1)
    def _draw():
        return set[selector()]
    return _draw

def exponential(minval, maxval, mean, limiter=redraw):
    def _draw():
        return random.expovariate(1.0/mean)
    return limiter(minval, maxval)(_draw)

def multimodal(weighted_distributions):
    total_weight = sum([w for (d, w) in weighted_distributions])
    selector = uniform(0, total_weight)
    def _draw():
        x = selector()
        wsum = 0
        for (d, w) in weighted_distributions:
            wsum += w
            if wsum >= x:
                return d()
        assert False # should never drop off
    return _draw

class TaskGenerator(object):
    def __init__(self, period, util, deadline=lambda x: x):
        self.period    = period
        self.util    = util
        self.deadline    = deadline

    def tasks(self, max_tasks=None, max_util=None, squeeze=False, tweak=None):
        count = 0
        usum  = 0
        while ((max_tasks is None or count < max_tasks) and
               (max_util is None  or usum  < max_util)):
            period   = self.period()
            util     = self.util()
            deadline = self.deadline(period)
            cost     = period * util
            if tweak:
                cost, period, deadline = tweak(cost, period, deadline)
                util = cost / float(period)
            count  += 1
            usum   += util
            if max_util and usum > max_util:
                if squeeze:
                    # make last task fit exactly
                    util -= (usum - max_util)
                    cost     = period * util
                else:
                    break
            yield ts.Task(cost, period, deadline)

    def task_sets(self, count, max_tasks, max_util, filter=None):
        while count > 0:
            set = ts.TaskSystem(self.tasks(max_tasks, max_util))
            set.index_by_period()
            if filter:
                _set = set
                set = filter(set)
                # special case: if filter returned True
                # then use unmodified set
                if set == True:
                    set = _set
            if set:
                count -= 1
                yield set

    def make_task_set(self, umax, max_tasks, *extra, **kextra):
        return ts.TaskSystem(self.tasks(max_tasks, umax, *extra, **kextra))

