from util import us2ms
from math import floor, ceil

def irq_dbf(cost, period, delta):
    times = int(floor(delta / float(period))) # full invocations
    last  = delta - times * period
    return times * cost + min(cost, last)

class InterruptDBF(object):
    "Demand bound function for sporadic interrupt."
    def __init__(self, cost, period):
        self.cost   = cost
        self.period = period
    
    def __call__(self, t):
        return irq_dbf(self.cost, self.period, t)

    def __str__(self):
        return "InterruptDBF(%s, %s)" % (self.cost, self.period)


def global_task_centric(lirq, girq, tasks):
    # lirq: list of list of local irq DBFs (per-processor)
    # girq: list of global irq DBFs
    for t in tasks:
        inf = 0
        for g in girq:
            inf += g(t.completion)
        for proc in lirq:
            for l in proc:
                inf += l(t.completion)
        t.cost += inf
    return tasks

def release_irqs_from_tasks(overheads, tasks):
    rel = us2ms(overheads.release(tasks))
    irqs = [InterruptDBF(rel, t.period) for t in tasks]
    return irqs

def tick_irqs_from_tasks(overheads, cpus, tasks):
    tick = us2ms(overheads.tick(tasks))
    return [[InterruptDBF(tick, us2ms(overheads.quantum_size))]
            for _ in xrange(cpus)]

def global_tick_only_task_centric(overheads, cpus, tasks):
    "Only account for ticks, not for releases."
    tic = us2ms(overheads.tick(tasks))
    ticp = us2ms(overheads.quantum_size)
#    print 'tic', tic, 'ticp', ticp
    for ti in tasks:
        n = floor(ti.period / float(ticp))
        ti.cost += (n + 2) * tic # add two extra for preemption
                                 # & migration surcharge (inflicted upon others)
    return tasks

def global_release_tick_task_centric(overheads, cpus, tasks):
    "Account for both release interrupts and tick overheads."
    rel = us2ms(overheads.release(tasks))
    tic = us2ms(overheads.tick(tasks))
    ticp = us2ms(overheads.quantum_size)
    for ti in tasks:
        rtotal = 0
        for tj in tasks:
            rtotal += irq_dbf(rel, tj.period, ti.completion)
        ti.cost += rtotal
        # see above for reasoning on tick charges
        n = floor(ti.period / float(ticp))
        ti.cost += (n + 2) * tic
    return tasks

def global_release_tick_task_centric_slow(overheads, cpus, tasks):
    "Nicer but slower version."
    lirq = tick_irqs_from_tasks(overheads, cpus, tasks)
    girq = release_irqs_from_tasks(overheads, tasks)
    return global_task_centric(lirq, girq, tasks)



def global_release_delay(overheads, tasks,
                         release_master=False,
                         inf_exec=False,
                         multiplexed=False):
    """Account for delayed releases in event-driven schedulers.
    Does not account for ticks, and also does not account for
    interference from release interrupts.
    """
    ipi_delay = us2ms(overheads.ipi_delay(tasks))
    irq_delay = us2ms(overheads.release_latency(tasks))

    delay = ipi_delay + irq_delay

    for t in tasks:
        t.cost += delay

    if release_master:
        # delay of release; with normal scheduling this is accounted for
        # as part of the IRQ charging in irq.py
        # add delay J to exec cost (see JSA'10 paper)

        if multiplexed:
            # if releases are multiplexed over a timer
            # then only two release ISRs can delay.
            rel_charge = us2ms(overheads.release(tasks)) * 2
        else:
            rel_charge = us2ms(overheads.release(tasks)) * len(tasks)

        if inf_exec:
            for t in tasks:
                t.cost += rel_charge
        else:
            for t in tasks:
                t.period -= rel_charge
