#!/usr/bin/env python

from model.tasks import TaskSystem, Task
from util import maybe_int, storage
import xml.etree.ElementTree as ET

def task(t):
    tag = ET.Element('task')
    if not t.id is None:
        tag.set('id', str(t.id))
    tag.set('period', str(maybe_int(t.period)))
    tag.set('wcet', str(maybe_int(t.cost)))
    if not t.implicit_deadline():
        tag.set('deadline', str(maybe_int(t.deadline)))
    if t.phase != 0:
        tag.set('phase', str(maybe_int(t.phase)))
    if not t.partition is None:
        tag.set('partition', str(t.partition))

    if t.resmodel:
        rmodel = ET.SubElement(tag, 'resources')
        for req in t.resmodel:
            if len(req) == 3:
             (res_id, req_num, req_len) = req
             kind = None
            else:
                (res_id, req_num, req_len, kind) = req
            req = ET.SubElement(rmodel, 'request')
            req.set('resid', str(res_id))
            req.set('count', str(req_num))
            req.set('length', str(req_len))
            if kind:
                req.set('type', str(kind))

    tag.task = t
    task.xml = tag
    return tag

def parse_request(req_node):
    return (int(req_node.get('resid')),
            int(req_node.get('count')),
            float(req_node.get('length')))

def parse_resmodel(node):
    resmodel = node.find('resources')
    if resmodel != None:
        return [parse_request(n) for n in resmodel.findall('request')]
    else:
        return []

def parse_task(node):
    cost      = float(node.get('wcet'))
    period    = float(node.get('period'))
    kargs = {}
    deadline  = node.get('deadline', None)
    if not deadline is None:
        kargs['deadline'] = float(deadline)
    phase     = node.get('phase', None)
    if not phase is None:
        kargs['phase'] = float(phase)
    id        = node.get('id', None)
    if not id is None:
        kargs['id'] = int(id)
    part = node.get('partition', None)
    if not part is None:
        kargs['partition'] = part

    t = Task(cost, period, **kargs)
    t.resmodel = parse_resmodel(node)

    t.xml = node
    node.task = t
    return t

def taskset(ts):
    tag = ET.Element('taskset')

    prop = ET.SubElement(tag, 'properties')
    prop.set('utilization', str(ts.utilization()))
    prop.set('density', str(ts.density()))
    hp = ts.hyperperiod()
    if hp:
        prop.set('hyperperiod', str(hp))

    for t in ts:
        tag.append(task(t))
    return tag

def testpoint(tasksets, params):
    tag = ET.Element('testpoint')

    config = ET.SubElement(tag, 'config')
    for k in params:
        config.set(k, str(params[k]))
    for ts in tasksets:
        tag.append(taskset(ts))
    return tag

def parse_taskset(node):
    tasks = [parse_task(n) for n in node.findall('task')]
    return TaskSystem(tasks)

def parse_testpoint(node):
    params = {}
    config = node.find('config')
    if not config is None:
        for k in config.keys():
            params[k] = config.get(k)
    tss = [parse_taskset(n) for n in node.findall('taskset')]
    return (params, tss)

def write(root, file):
    tree = ET.ElementTree(root)
    tree.write(file)

def write_testpoint(tasksets, params, fname):
    xml = testpoint(tasksets, params)
    write(xml, fname)

def write_ts(ts, file):
    xml = taskset(ts)
    write(xml, file)

def load(file):
    tree = ET.ElementTree()
    tree.parse(file)
    root = tree.getroot()
    if root.tag == 'taskset':
        ts = parse_taskset(root)
        ts.xml = tree.getroot()
        return ts
    elif root.tag == 'testpoint':
        params, tss = parse_testpoint(root)
        return storage(xml=root, params=params, tasksets=tss)
    else:
        return None


