"""Support for quantum-based scheduling.
"""

from math      import ceil, floor
from util      import is_multiple

def has_integral_period(qlen):
    return lambda t: is_multiple(t.period, qlen)

def quantize_wcet(qlen, tasks, effective_qlen=None):
    """Round up execution cost to account for partially used quanta.
    Specify an effective_qlen less than the quantum length to account for
    overheads.
    """
    if effective_qlen is None:
        effective_qlen = qlen
    assert effective_qlen > 0
    assert qlen > 0
    for t in tasks:
        nr_quanta = int(ceil(t.cost / float(effective_qlen)))
        t.cost = nr_quanta * qlen
    return tasks

def quantize_period(qlen, tasks):
    """Round down periods to account for the fact that in a quantum-based
    scheduler all periods must be multiples of the quantum length.

    Rounding down the period of a periodic task yields a sporadic task that has
    an interarrival delay of one quantum.
    """
    for t in tasks:
        if not is_multiple(t.period, qlen):
            nr_quanta = int(floor(t.period / float(qlen)))
            per = nr_quanta * qlen
            if t.implicit_deadline():
                # implicit deadline => also update deadline
                t.deadline = per
            t.period = per
    return tasks

def account_for_delayed_release(delay, tasks):
    """A release will not be noticed until the start of the next quantum
    boundary. Hence, the period must be reduced by one quantum size for hard
    real-time use.
    """
    for t in tasks:
        if t.implicit_deadline():
            t.deadline -= delay
        t.period -= delay
    return tasks

def account_for_staggering(qlen, no_cpus, tasks):
    """A job may miss its deadline by up to ((m - 1) / m) of a quantum length
    due to staggering. Hence, we need to reduce the period.
    FIXME: What to do if qlen is int?
    """
    m    = no_cpus
    miss = (m - 1) / float(m)
    reduction = miss * qlen
    for t in tasks:
        if t.implicit_deadline():
            t.deadline -= reduction
        t.period -= reduction
    return tasks
